"""Container for event loop. Prevent loop to exit when no watchers exists."""
from __future__ import absolute_import

import logging

from threading import Event
from pyuv import Async

from .utils.threads import spawn
from .utils.loop import in_loop
from .utils.mixin import LoopMixin

__all__ = ['LoopContainer']

logger = logging.getLogger(__name__)


class LoopContainer(LoopMixin):
    """Container for event loop."""

    app = None

    def __init__(self):
        self._guard_watcher = None
        self._started = Event()
        self._stopped = Event()

    def _run(self):
        """Run event loop."""
        self._started.set()
        loop = self.loop
        logger.debug('Loop %r started...', loop)
        try:
            loop.run()
            logger.debug('Loop %r stopped...', loop)
            self._stopped.set()
        except Exception as exc:
            logger.exception(exc)

    @in_loop
    def _close_handlers(self):
        """Close all stale handlers."""
        def cb_handle(handle):
            if not handle.closed:
                logger.warning('Close stale handle %r', handle)
                handle.close()
        self.loop.walk(cb_handle)

    def start(self):
        """Start event loop in separate thread."""
        self._guard_watcher = Async(self.loop, lambda *args: None)
        spawn(self._run)
        self._started.wait()

    def stop(self):
        """Stop event loop and wait until it exit."""
        assert self._guard_watcher is not None, 'loop not started'
        self._guard_watcher.close()
        self._close_handlers()
        self._stopped.wait()



"""Some other useful tools."""
from __future__ import absolute_import

import itertools

__all__ = ['get_port_from_range', 'get_addresses_from_pool',
           'rgetattr']


def rgetattr(obj, path):
    """Get nested attribute from object.

    :param obj: object
    :param path: path to attribute

    """
    return reduce(getattr, [obj] + path.split('.'))


def get_port_from_range(name, range_start, range_end):
    """Get random port from given range [range_start, range_end]."""
    # Detect port for this service.
    hashed_port = abs(hash(name) & 0xffff)
    # Ensure that detected port is in pool.
    while hashed_port > range_start:
        hashed_port = hashed_port - range_start
    # If port is greater than allowed.
    while range_end < hashed_port + range_start:
        hashed_port = hashed_port // 2
    return hashed_port + range_start


def get_addresses_from_pool(name, address, port_range):
    """Get addresses from pool."""
    host, port = address
    if isinstance(port, int):
        ports = (port or 0,)
    elif isinstance(port, basestring) or port is None:
        if port_range is None and port is not None:
            raise ValueError('Port range not specified')
        elif port == 'random':
            ports = (0,)
        elif port is None or port == 'pool':
            range_start, range_end = port_range[0], port_range[1]
            default_port = get_port_from_range(name, range_start, range_end)
            chains = [(default_port,),
                      xrange(default_port, range_end),
                      xrange(range_start, default_port)]
            ports = itertools.chain.from_iterable(chains)
        elif port.isdigit():
            ports = (int(port),)
        else:
            raise ValueError('Unknown port {0!r}'.format(port))
    else:
        raise ValueError('Unknown address {0!r}'.format(address))
    for port in ports:
        yield (host, port)



from logging import getLogger
from struct import Struct
from io import BytesIO

cimport cython

from pyuv.errno import strerror, UV_EOF

from .constants import LENGTH_FORMAT, LENGTH_SIZE, NONBLOCKING

__all__ = ['Connection']

logger = getLogger(__name__)


cdef class Connection:
    """Basic class is represented connection.

    It can be in state:
        WAIT_LEN --- connection is reading request length.
        WAIT_MESSAGE --- connection is reading request.
        WAIT_PROCESS --- connection has just read whole request and
            waits for call ready routine.
        SEND_ANSWER --- connection is sending answer string (including length
            of answer).
        CLOSED --- socket was closed and connection should be deleted.

    """

    def __init__(self, object producer, object loop, object client,
                 object on_close):
        # Default values.
        self.recv_bytes = self.message_length = 0
        self.status = WAIT_LEN
        self.struct = Struct(LENGTH_FORMAT)
        self.message_buffer = BytesIO()
        self.incoming_buffer = None

        # Given arguments.
        self.producer = producer
        self.loop = loop
        self.client = client
        self.on_close = on_close

        # Start watchers.
        self.peername = '{0[0]}:{0[1]}'.format(self.client.getpeername())
        self.client.start_read(self.cb_read_done)

    @cython.profile(False)
    cdef inline bint is_writeable(self):
        """Returns ``True`` if source is writable."""
        return self.status == SEND_ANSWER

    @cython.profile(False)
    cdef inline bint is_readable(self):
        """Returns ``True`` if source is readable."""
        return self.status == WAIT_LEN or self.status == WAIT_MESSAGE

    @cython.profile(False)
    cdef inline bint is_ready(self):
        """Returns ``True`` if source is ready."""
        return self.status == WAIT_PROCESS

    @cython.profile(False)
    cpdef is_waiting(self):
        """Returns ``True`` if source is waiting for answer."""
        return self.status == WAIT_ANSWER

    @cython.profile(False)
    cpdef is_closed(self):
        """Returns ``True`` if source is closed."""
        return self.status == CLOSED

    cdef inline read(self):
        """Reads data from stream and switch state."""
        assert self.is_readable(), 'socket in non-readable state'
        received = len(self.incoming_buffer)

        if received == 0:
            # if we read 0 bytes and message is empty, it means client
            # close connection
            self.close()
            return

        if self.status == WAIT_LEN:
            assert received > LENGTH_SIZE, "message length can't be read"

            view = buffer(self.incoming_buffer)
            message_length = self.struct.unpack_from(view[0:LENGTH_SIZE])[0]
            assert message_length > 0, "negative or empty frame size, it seems" \
                                       " client doesn't use FramedTransport"

            self.message_buffer.write(view[LENGTH_SIZE:])
            self.message_length = message_length
            self.status = WAIT_MESSAGE

            received = received - LENGTH_SIZE

        elif self.status == WAIT_MESSAGE:
            # Simply write data to buffer.
            self.message_buffer.write(self.incoming_buffer)

        self.recv_bytes += received
        if self.recv_bytes == self.message_length:
            self.recv_bytes = 0
            self.status = WAIT_PROCESS

    cpdef close(self):
        """Closes connection."""
        assert not self.is_closed(), 'socket already closed'

        # Close socket.
        self.status = CLOSED
        self.client.close()

        # Execute callback.
        self.on_close(self)

        # Remove references to objects.
        self.on_close = self.client = self.message_buffer = \
            self.incoming_buffer = None

    cpdef ready(self, object all_ok, object message):
        """The ready can switch Connection to three states:

            WAIT_LEN if request was oneway.
            SEND_ANSWER if request was processed in normal way.
            CLOSED if request throws unexpected exception.

        """
        assert self.is_waiting(), 'socket is not waiting for answer'

        if not all_ok:
            self.close()
            return

        message_length = self.message_length = len(message)

        if message_length == 0:
            # it was a oneway request, do not write answer
            self.status = WAIT_LEN
        else:
            # Create message.
            data = self.struct.pack(message_length)
            data += message
            self.status = SEND_ANSWER
            # Write data.
            self.client.write(data, self.cb_write_done)

    cdef inline void handle_error(self, object error):
        logger.error('Error with client %r: %s',
                     self.peername, strerror(error))

    cpdef cb_read_done(self, object handle, object data, object error):
        if error and error != UV_EOF:
            self.handle_error(error)
            self.close()
            return

        try:
            self.incoming_buffer = data or ''
            while self.is_readable():
                # Try to read whole message to buffer while we can.
                self.read()

            if not self.is_ready():
                # We aren't ready to transfer message to sink.
                return
            elif not self.is_waiting():
                # Change state to needed.
                self.status = WAIT_ANSWER
                # Send message to workers.
                self.producer(self, self.message_buffer.getvalue())
                # Reset message buffer.
                self.message_buffer = BytesIO()
            else:
                # Socket was closed while we wait for answer.
                self.close()

        except Exception as exc:
            logger.exception(exc)
            self.close()

    cpdef cb_write_done(self, object handle, object error):
        assert self.is_writeable(), 'socket in non writable state'

        if error:
            self.handle_error(error)
            self.close()
            return

        self.status = WAIT_LEN



"""Process incoming request and return it's result."""

VERSION = (0, 1, 6)

__version__ = '.'.join(map(str, VERSION[0:3]))
__author__ = 'Lipin Dmitriy'
__contact__ = 'blackwithwhite666@gmail.com'
__homepage__ = 'https://github.com/blackwithwhite666/thriftworker'
__docformat__ = 'restructuredtext'

# -eof meta-

from .app import ThriftWorker



"""Detect current environment."""
from __future__ import absolute_import

import sys

from thriftworker.constants import DEFAULT_ENV, GEVENT_ENV

__all__ = ['detect_environment']

_environment = None


def _detect_environment():

    # -gevent-
    if 'gevent' in sys.modules:
        try:
            from gevent import socket as _gsocket
            import socket

            if socket.socket is _gsocket.socket:
                return GEVENT_ENV
        except ImportError:
            pass

    return DEFAULT_ENV


def detect_environment():
    global _environment
    if _environment is None:
        _environment = _detect_environment()
    return _environment



"""Handle global package state."""
from __future__ import absolute_import

from .utils.proxy import Proxy

default_app = None


def set_current_app(app):
    global default_app
    default_app = app


def get_current_app():
    global default_app
    if default_app is None:
        # creates the default app, but we want to defer that.
        raise RuntimeError('No ThriftWorker was configured!')
    return default_app


current_app = Proxy(get_current_app)



"""SocketZMQ.

Distribute thrift requests between workers.

"""
from __future__ import absolute_import

from pyuv import Loop
from thrift.protocol import TBinaryProtocol

from .state import set_current_app, get_current_app
from .listener import Listener
from .loop import LoopContainer
from .services import Services
from .worker import Worker
from .utils.decorators import cached_property
from .utils.mixin import SubclassMixin


class ThriftWorker(SubclassMixin):
    """Store global application state. Also acts as factory for whole
    application.

    """

    def __init__(self, protocol_factory=None, port_range=None):
        # Set provided instance if we can.
        if protocol_factory is not None:
            self.protocol_factory = protocol_factory
        self.port_range = port_range
        super(ThriftWorker, self).__init__()
        set_current_app(self)

    @staticmethod
    def instance():
        """Return global instance."""
        return get_current_app()

    @cached_property
    def loop(self):
        """Create event loop. Should be running in separate thread."""
        return Loop()

    @cached_property
    def LoopContainer(self):
        """Create bounded :class:`LoopContainer` class."""
        return self.subclass_with_self(LoopContainer)

    @cached_property
    def loop_container(self):
        """Instance of bounded :class:`LoopContainer`."""
        return self.LoopContainer()

    @cached_property
    def protocol_factory(self):
        """Specify which protocol should be used."""
        return TBinaryProtocol.TBinaryProtocolAcceleratedFactory()

    @cached_property
    def Services(self):
        """Create bounded :class:`Processor` class."""
        return self.subclass_with_self(Services)

    @cached_property
    def services(self):
        """Create global request processor instance."""
        return self.Services()

    @cached_property
    def Worker(self):
        """Create bounded pool class."""
        return self.subclass_with_self(Worker)

    @cached_property
    def worker(self):
        """Create pool."""
        return self.Worker()

    @cached_property
    def Listener(self):
        """Create bounded :class:`Listener` class."""
        return self.subclass_with_self(Listener)



# -*- coding: utf-8 -*-
"""Parts of this module is Copyright by Werkzeug Team."""
from __future__ import absolute_import


class Proxy(object):
    """Proxy to another object."""

    # Code stolen from werkzeug.local.Proxy.
    __slots__ = ('__local', '__args', '__kwargs', '__dict__', '__name__')

    def __init__(self, local, args=None, kwargs=None, name=None):
        object.__setattr__(self, '_Proxy__local', local)
        object.__setattr__(self, '_Proxy__args', args or ())
        object.__setattr__(self, '_Proxy__kwargs', kwargs or {})
        if name is not None:
            object.__setattr__(self, '__custom_name__', name)

    @property
    def __name__(self):
        try:
            return self.__custom_name__
        except AttributeError:
            return self._get_current_object().__name__

    @property
    def __module__(self):
        return self._get_current_object().__module__

    @property
    def __doc__(self):
        return self._get_current_object().__doc__

    def _get_class(self):
        return self._get_current_object().__class__

    @property
    def __class__(self):
        return self._get_class()

    def _get_current_object(self):
        """Return the current object.  This is useful if you want the real
        object behind the proxy at a time for performance reasons or because
        you want to pass the object into a different context.
        """
        if not hasattr(self.__local, '__release_local__'):
            return self.__local(*self.__args, **self.__kwargs)
        try:
            return getattr(self.__local, self.__name__)
        except AttributeError:
            raise RuntimeError('no object bound to {0.__name__}'.format(self))

    @property
    def __dict__(self):
        try:
            return self._get_current_object().__dict__
        except RuntimeError:  # pragma: no cover
            raise AttributeError('__dict__')

    def __repr__(self):
        try:
            obj = self._get_current_object()
        except RuntimeError:  # pragma: no cover
            return '<{0} unbound>'.format(self.__class__.__name__)
        return repr(obj)

    def __nonzero__(self):
        try:
            return bool(self._get_current_object())
        except RuntimeError:  # pragma: no cover
            return False

    def __unicode__(self):
        try:
            return unicode(self._get_current_object())
        except RuntimeError:  # pragma: no cover
            return repr(self)

    def __dir__(self):
        try:
            return dir(self._get_current_object())
        except RuntimeError:  # pragma: no cover
            return []

    def __getattr__(self, name):
        if name == '__members__':
            return dir(self._get_current_object())
        return getattr(self._get_current_object(), name)

    def __setitem__(self, key, value):
        self._get_current_object()[key] = value

    def __delitem__(self, key):
        del self._get_current_object()[key]

    def __setslice__(self, i, j, seq):
        self._get_current_object()[i:j] = seq

    def __delslice__(self, i, j):
        del self._get_current_object()[i:j]

    __setattr__ = lambda x, n, v: setattr(x._get_current_object(), n, v)
    __delattr__ = lambda x, n: delattr(x._get_current_object(), n)
    __str__ = lambda x: str(x._get_current_object())
    __lt__ = lambda x, o: x._get_current_object() < o
    __le__ = lambda x, o: x._get_current_object() <= o
    __eq__ = lambda x, o: x._get_current_object() == o
    __ne__ = lambda x, o: x._get_current_object() != o
    __gt__ = lambda x, o: x._get_current_object() > o
    __ge__ = lambda x, o: x._get_current_object() >= o
    __cmp__ = lambda x, o: cmp(x._get_current_object(), o)
    __hash__ = lambda x: hash(x._get_current_object())
    __call__ = lambda x, *a, **kw: x._get_current_object()(*a, **kw)
    __len__ = lambda x: len(x._get_current_object())
    __getitem__ = lambda x, i: x._get_current_object()[i]
    __iter__ = lambda x: iter(x._get_current_object())
    __contains__ = lambda x, i: i in x._get_current_object()
    __getslice__ = lambda x, i, j: x._get_current_object()[i:j]
    __add__ = lambda x, o: x._get_current_object() + o
    __sub__ = lambda x, o: x._get_current_object() - o
    __mul__ = lambda x, o: x._get_current_object() * o
    __floordiv__ = lambda x, o: x._get_current_object() // o
    __mod__ = lambda x, o: x._get_current_object() % o
    __divmod__ = lambda x, o: x._get_current_object().__divmod__(o)
    __pow__ = lambda x, o: x._get_current_object() ** o
    __lshift__ = lambda x, o: x._get_current_object() << o
    __rshift__ = lambda x, o: x._get_current_object() >> o
    __and__ = lambda x, o: x._get_current_object() & o
    __xor__ = lambda x, o: x._get_current_object() ^ o
    __or__ = lambda x, o: x._get_current_object() | o
    __div__ = lambda x, o: x._get_current_object().__div__(o)
    __truediv__ = lambda x, o: x._get_current_object().__truediv__(o)
    __neg__ = lambda x: -(x._get_current_object())
    __pos__ = lambda x: +(x._get_current_object())
    __abs__ = lambda x: abs(x._get_current_object())
    __invert__ = lambda x: ~(x._get_current_object())
    __complex__ = lambda x: complex(x._get_current_object())
    __int__ = lambda x: int(x._get_current_object())
    __long__ = lambda x: long(x._get_current_object())
    __float__ = lambda x: float(x._get_current_object())
    __oct__ = lambda x: oct(x._get_current_object())
    __hex__ = lambda x: hex(x._get_current_object())
    __index__ = lambda x: x._get_current_object().__index__()
    __coerce__ = lambda x, o: x.__coerce__(x, o)
    __enter__ = lambda x: x._get_current_object().__enter__()
    __exit__ = lambda x, *a, **kw: x._get_current_object().__exit__(*a, **kw)
    __reduce__ = lambda x: x._get_current_object().__reduce__()


class PromiseProxy(Proxy):
    """This is a proxy to an object that has not yet been evaluated.

    :class:`Proxy` will evaluate the object each time, while the
    promise will only evaluate it once.

    """

    def _get_current_object(self):
        try:
            return object.__getattribute__(self, '__thing')
        except AttributeError:
            return self.__evaluate__()

    def __evaluated__(self):
        try:
            object.__getattribute__(self, '__thing')
        except AttributeError:
            return False
        return True

    def __maybe_evaluate__(self):
        return self._get_current_object()

    def __evaluate__(self):
        try:
            thing = Proxy._get_current_object(self)
            object.__setattr__(self, '__thing', thing)
            return thing
        finally:
            object.__delattr__(self, '_Proxy__local')
            object.__delattr__(self, '_Proxy__args')
            object.__delattr__(self, '_Proxy__kwargs')



"""Base pool implementation."""
from __future__ import absolute_import

import logging
from collections import namedtuple

from pyuv import ThreadPool

from .utils.decorators import cached_property

logger = logging.getLogger(__name__)


class Worker(object):
    """Process all request in separate thread."""

    app = None

    #: Store request in this tuple.
    Request = namedtuple('Request', 'connection data')

    @cached_property
    def pool(self):
        return ThreadPool(self.app.loop)

    def create_callback(self, request):
        """Create callback that should be called after request was done."""
        connection = request.connection

        def inner_callback(result, exception):
            if exception is None:
                success, response = True, result
            else:
                logger.error(exception)
                success, response = False, ''
            connection.ready(success, response)

        return inner_callback

    def create_consumer(self, processor):
        create_callback = self.create_callback
        pool = self.pool

        def inner_consumer(request):
            # Nested function that process incoming request.
            task = lambda: processor(request.data)
            # Put task to pool.
            pool.queue_work(task, create_callback(request))

        return inner_consumer

    def create_producer(self, service):
        """Create producer for connections."""
        Request = self.Request
        processor = self.app.services.create_processor(service)
        consume = self.create_consumer(processor)

        def inner_producer(connection, data):
            """Enqueue given request to thread pool."""
            consume(Request(connection=connection, data=data))

        return inner_producer



"""Some function to work with system threads."""
from __future__ import absolute_import

import imp
import os
import sys
import traceback
from threading import Thread, Event

__all__ = ['spawn', 'DaemonThread']

_realthread = None


def get_realthread():
    """Get the real Python thread module, regardless of any monkeypatching"""
    global _realthread
    if _realthread:
        return _realthread

    fp, pathname, description = imp.find_module('thread')
    try:
        _realthread = imp.load_module('realthread', fp, pathname, description)
        return _realthread
    finally:
        if fp:
            fp.close()


def spawn(func, *args, **kwargs):
    """Takes a function and spawns it as a daemon thread using the
    real OS thread regardless of monkey patching.

    """
    return get_realthread().start_new_thread(func, args, kwargs)


class DaemonThread(Thread):

    def __init__(self, name=None, **kwargs):
        super(DaemonThread, self).__init__()
        self._is_stopped = Event()
        self.daemon = True
        self.name = name or self.__class__.__name__

    def body(self):
        raise NotImplementedError('subclass responsibility')

    def run(self):
        try:
            self.body()
        except Exception as exc:
            try:
                self.on_crash('{0!r} crashed: {1!r}', self.name, exc)
            finally:
                os._exit(1)  # exiting by normal means won't work
        finally:
            self._set_stopped()

    def _set_stopped(self):
        try:
            self._is_stopped.set()
        except TypeError:  # pragma: no cover
            # we lost the race at interpreter shutdown,
            # so gc collected built-in modules.
            pass

    def stop(self):
        """Graceful shutdown."""
        self._is_stopped.wait()
        if self.is_alive():
            self.join(1e100)

    def on_crash(self, msg, *fmt):
        sys.stderr.write(msg.format(*fmt) + '\n')
        exc_info = sys.exc_info()
        try:
            traceback.print_exception(exc_info[0], exc_info[1], exc_info[2],
                                      None, sys.stderr)
        finally:
            del(exc_info)


class LoopThread(DaemonThread):

    def __init__(self, name=None, **kwargs):
        super(LoopThread, self).__init__(name, **kwargs)
        self._is_shutdown = Event()

    def on_start(self):
        pass

    def on_stop(self):
        pass

    def body(self):
        self.on_start()
        while not self._is_shutdown.is_set():
            self.loop()
        self.on_stop()

    def loop(self):
        raise NotImplementedError('subclass responsibility')

    def stop(self):
        """Graceful shutdown."""
        self._is_shutdown.set()
        super(LoopThread, self).stop()


class SimpleDaemonThread(DaemonThread):

    def __init__(self, target, name=None, **kwargs):
        self.target = target
        super(SimpleDaemonThread, self).__init__(name, **kwargs)

    def body(self):
        self.target()



"""Combine device and proxy together. Workers must connect to device backend.

"""
from __future__ import absolute_import

import logging

from pyuv import TCP
from pyuv.error import TCPError
from pyuv.errno import UV_EADDRINUSE, strerror

from .constants import BACKLOG_SIZE
from .exceptions import BindError
from .connection import Connection
from .utils.loop import in_loop
from .utils.decorators import cached_property
from .utils.other import get_addresses_from_pool
from .utils.mixin import LoopMixin

__all__ = ['Listener']

logger = logging.getLogger(__name__)


class Connections(object):
    """Store connections."""

    def __init__(self):
        self.connections = set()

    def register(self, connection):
        """Register new connection."""
        self.connections.add(connection)

    def remove(self, connection):
        """Remove registered connection."""
        try:
            self.connections.remove(connection)
        except KeyError:
            logger.warning('Connection %r not registered', connection)

    def close(self):
        while self.connections:
            connection = self.connections.pop()
            if not connection.is_closed():
                connection.close()


class Listener(LoopMixin):
    """Facade for proxy. Support lazy initialization."""

    app = None

    Connections = Connections

    def __init__(self, name, address, backlog=None):
        """Create new listener.

        :param name: service name
        :param address: address of socket
        :param backlog: size of socket connection queue

        """
        self.name = name
        self.address = address
        self.backlog = backlog or BACKLOG_SIZE
        self._connections = self.Connections()

    def _create_socket(self):
        return TCP(self.loop)

    @cached_property
    def _socket(self):
        """A shortcut to create a TCP socket and bind it."""
        return self._create_socket()

    @property
    def host(self):
        """Return host to which this socket is binded."""
        return self._socket.getsockname()[0]

    @property
    def port(self):
        """Return binded port number."""
        return self._socket.getsockname()[1]

    def _create_acceptor(self):
        service = self.name
        loop = self.loop
        connections = self._connections
        producer = self.app.worker.create_producer(service)
        server_socket = self._socket
        socket_factory = self._create_socket

        def on_close(connection):
            connections.remove(connection)

        def on_connection(handle, error):
            if error:
                logger.error('Error handling new connection for service %r: %s',
                             service, strerror(error))
                return
            client = socket_factory()
            client.nodelay(True)
            server_socket.accept(client)
            connection = Connection(producer, loop, client, on_close)
            connections.register(connection)

        return on_connection

    @in_loop
    def start(self):
        binded = False
        socket = self._socket
        acceptor = self._create_acceptor()
        backlog = self.backlog
        addresses = get_addresses_from_pool(self.name, self.address,
                                            self.app.port_range)
        for address in addresses:
            try:
                socket.bind(address)
                socket.listen(acceptor, backlog)
            except TCPError as exc:
                if exc.args[0] == UV_EADDRINUSE:
                    continue
                raise
            else:
                binded = True
                break
        if not binded:
            raise BindError("Service {0!r} can't bind to address {1!r}".
                            format(self.name, self.address))

    @in_loop
    def stop(self):
        self._socket.close()
        self._connections.close()



"""All exceptions for this package."""
from __future__ import absolute_import


class AllocationError(Exception):
    """Raised if we can't bind to any address from pool."""


class BindError(Exception):
    """Error on socket binding."""


class ServiceAlreadyRegistered(Exception):
    """Given service already registered in storage."""



"""Some useful mixins here."""
from __future__ import absolute_import

from ..state import current_app
from .other import rgetattr

__all__ = ['LoopMixin', 'SubclassMixin', 'PropertyMixin']


class LoopMixin(object):
    """Move loop closer to instance."""

    app = None

    @property
    def loop(self):
        """Shortcut to loop."""
        return self.app.loop


def _unpickle_appattr(name, args):
    """Given an attribute name and a list of args, gets
    the attribute from the current app and calls it.

    """
    return rgetattr(current_app, name)(*args)


class SubclassMixin(object):

    def subclass_with_self(self, Class, name=None, attribute='app',
                           reverse=None, **kw):
        """Subclass an app-compatible class by setting its app attribute
        to be this app instance.

        App-compatible means that the class has a class attribute that
        provides the default app it should use, e.g.
        ``class Foo: app = None``.

        :param Class: The app-compatible class to subclass.
        :keyword name: Custom name for the target class.
        :keyword attribute: Name of the attribute holding the app,
                            default is 'app'.

        """
        reverse = reverse if reverse else Class.__name__

        def __reduce__(self):
            return (_unpickle_appattr,
                    (Class.__name__, self.__reduce_args__()))

        attrs = dict({attribute: self},
                     __module__=Class.__module__,
                     __reduce__=__reduce__,
                     __doc__=Class.__doc__, **kw)

        return type(name or Class.__name__, (Class,), attrs)


def _unpickle_appprop(name):
    """Given an attribute name, gets the attribute from the current app."""
    return rgetattr(current_app, name)


class PropertyMixin(object):

    def __reduce__(self):
        return (_unpickle_appprop, (self.__class__.__name__.lower(),))



"""Store processor and protocol for each service."""
from __future__ import absolute_import

from collections import namedtuple

from thrift.transport.TTransport import TMemoryBuffer

from .exceptions import ServiceAlreadyRegistered


class Services(object):
    """Process new requests and return response. Store processor
    for each service.

    """

    app = None

    #: Holder of service processor and protocol factory.
    Service = namedtuple('Service', 'processor proto_factory')

    def __init__(self):
        self.services = {}
        self.proto_factory = self.app.protocol_factory
        super(Services, self).__init__()

    def register(self, service_name, processor, proto_factory=None):
        """Register new processor for given service."""
        service = self.Service(processor, proto_factory or self.proto_factory)
        if service_name in self.services:
            raise ServiceAlreadyRegistered('Service {0} already registered'
                                           .format(service_name))
        self.services[service_name] = service

    def create_processor(self, service_name):
        """Create function that will process incoming request and return
        payload that we should return.

        :param service_name: name of served service

        """
        service = self.services[service_name]
        proto_factory, processor = service.proto_factory, service.processor

        def inner_processor(payload):
            in_transport = TMemoryBuffer(payload)
            out_transport = TMemoryBuffer()
            in_prot = proto_factory.getProtocol(in_transport)
            out_prot = proto_factory.getProtocol(out_transport)
            processor.process(in_prot, out_prot)
            return out_transport.getvalue()

        return inner_processor



import os
import re
import sys

from setuptools import setup, Extension, find_packages
from setuptools.command.sdist import sdist
from setuptools.command.build_ext import build_ext


if sys.version_info < (2, 7):
    raise Exception('ThriftWorker requires Python 2.7.')


#-----------------------------------------------------------------------------
# Flags and default values.
#-----------------------------------------------------------------------------

cmdclass = {}
extensions = []
extension_kwargs = {}


# try to find cython
try:
    from Cython.Distutils import build_ext as build_ext_c
    cython_installed = True
except ImportError:
    cython_installed = False


#-----------------------------------------------------------------------------
# Commands
#-----------------------------------------------------------------------------

class CheckSDist(sdist):
    """Custom sdist that ensures Cython has compiled all pyx files to c."""

    def initialize_options(self):
        sdist.initialize_options(self)
        self._pyxfiles = []
        for root, dirs, files in os.walk('thriftworker'):
            for f in files:
                if f.endswith('.pyx'):
                    self._pyxfiles.append(os.path.join(root, f))

    def run(self):
        if 'cython' in cmdclass:
            self.run_command('cython')
        else:
            for pyxfile in self._pyxfiles:
                cfile = pyxfile[:-3] + 'c'
                msg = "C-source file '%s' not found."%(cfile)+\
                " Run 'setup.py cython' before sdist."
                assert os.path.isfile(cfile), msg
        sdist.run(self)

cmdclass['sdist'] = CheckSDist


if cython_installed:

    class CythonCommand(build_ext_c):
        """Custom distutils command subclassed from Cython.Distutils.build_ext
        to compile pyx->c, and stop there. All this does is override the
        C-compile method build_extension() with a no-op."""

        description = "Compile Cython sources to C"

        def build_extension(self, ext):
            pass

    class zbuild_ext(build_ext_c):
        def run(self):
            from distutils.command.build_ext import build_ext as _build_ext
            return _build_ext.run(self)

    cmdclass['cython'] = CythonCommand
    cmdclass['build_ext'] = zbuild_ext

else:

    class CheckingBuildExt(build_ext):
        """Subclass build_ext to get clearer report if Cython is neccessary."""

        def check_cython_extensions(self, extensions):
            for ext in extensions:
                for src in ext.sources:
                    msg = "Cython-generated file '%s' not found." % src
                    assert os.path.exists(src), msg

        def build_extensions(self):
            self.check_cython_extensions(self.extensions)
            self.check_extensions_list(self.extensions)

            for ext in self.extensions:
                self.build_extension(ext)

    cmdclass['build_ext'] = CheckingBuildExt


#-----------------------------------------------------------------------------
# Extensions
#-----------------------------------------------------------------------------


suffix = '.pyx' if cython_installed else '.c'


def source_extension(name):
    return os.path.join('thriftworker', name + suffix)


# collect extensions
for module in ['connection']:
    sources = [source_extension(module)]
    ext = Extension('thriftworker.{0}'.format(module),
                    sources=sources,
                    **extension_kwargs)
    if suffix == '.pyx' and ext.sources[0].endswith('.c'):
        # undo setuptools stupidly clobbering cython sources:
        ext.sources = sources
    extensions.append(ext)


#-----------------------------------------------------------------------------
# Description, version and other meta information.
#-----------------------------------------------------------------------------

re_meta = re.compile(r'__(\w+?)__\s*=\s*(.*)')
re_vers = re.compile(r'VERSION\s*=\s*\((.*?)\)')
re_doc = re.compile(r'^"""(.+?)"""')
rq = lambda s: s.strip("\"'")


def add_default(m):
    attr_name, attr_value = m.groups()
    return ((attr_name, rq(attr_value)),)


def add_version(m):
    v = list(map(rq, m.groups()[0].split(', ')))
    return (('VERSION', '.'.join(v[0:3]) + ''.join(v[3:])),)


def add_doc(m):
    return (('doc', m.groups()[0]),)

pats = {re_meta: add_default,
        re_vers: add_version,
        re_doc: add_doc}
here = os.path.abspath(os.path.dirname(__file__))
meta_fh = open(os.path.join(here, 'thriftworker/__init__.py'))
try:
    meta = {}
    for line in meta_fh:
        if line.strip() == '# -eof meta-':
            break
        for pattern, handler in pats.items():
            m = pattern.match(line.strip())
            if m:
                meta.update(handler(m))
finally:
    meta_fh.close()


#-----------------------------------------------------------------------------
# Setup
#-----------------------------------------------------------------------------

setup(
    name='thriftworker',
    version=meta['VERSION'],
    description=meta['doc'],
    author=meta['author'],
    author_email=meta['contact'],
    url=meta['homepage'],
    license='BSD',
    cmdclass=cmdclass,
    ext_modules=extensions,
    packages=find_packages(),
    install_requires=['pyuv', 'thrift'],
    classifiers=["Development Status :: 4 - Beta",
                 "Intended Audience :: Developers",
                 "Intended Audience :: System Administrators",
                 "License :: OSI Approved :: BSD License",
                 "Operating System :: POSIX",
                 "Programming Language :: Python :: 2.7"],
)



"""All constants for this package."""
from __future__ import absolute_import

import errno
from struct import calcsize


DEFAULT_ENV = 0x1
GEVENT_ENV = 0x2

LENGTH_FORMAT = '!i'

LENGTH_SIZE = calcsize(LENGTH_FORMAT)

BACKLOG_SIZE = 1024

NONBLOCKING = (errno.EAGAIN, errno.EWOULDBLOCK)



"""Execute some function in loop and wait for answer."""
from __future__ import absolute_import

import sys
from threading import Event
from functools import wraps

from pyuv import Async


class in_loop(object):
    """Schedule execution of given function in main event loop. Wait for
    function execution.

    """

    def __init__(self, func=None, timeout=None):
        self.__func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__
        self.__timeout = timeout or 5.0

    def __create(self, obj):
        method = self.__func.__get__(obj)

        @wraps(self.__func)
        def inner_decorator(*args, **kwargs):
            event = Event()
            d = {'result': None, 'tb': None}

            def inner_callback(async_handle):
                async.close()
                try:
                    d['result'] = method(*args, **kwargs)
                except Exception as exc:
                    # Save traceback here.
                    d['result'] = exc
                    d['tb'] = sys.exc_info()[2]
                finally:
                    event.set()

            async = Async(obj.loop, inner_callback)
            async.send()
            try:
                if not event.wait(self.__timeout):
                    raise Exception('Timeout happened when calling method'
                                    ' {0!r} of {1!r}'.format(self.__name__, obj))
            finally:
                event.clear()
                if not async.closed:
                    async.close()

            result, tb = d['result'], d['tb']
            if isinstance(result, Exception):
                # Restore traceback.
                raise result.__class__, result, tb
            else:
                return result

        return inner_decorator

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        try:
            return obj.__dict__[self.__name__]
        except KeyError:
            value = obj.__dict__[self.__name__] = self.__create(obj)
            return value
